/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type FieldHasDefault, type ImplicitAllowedTypes, type ImplicitFieldSchema, type InsertableTreeNodeFromImplicitAllowedTypes, type NodeKind, type SchemaFactoryAlpha, type ScopedSchemaName, TreeArrayNode, type TreeNode, type TreeNodeFromImplicitAllowedTypes, type TreeNodeSchemaClass, type WithType, type TreeFieldFromImplicitField, type InsertableTreeFieldFromImplicitField, type InternalTreeNode, type FieldSchema, type FieldKind, type ImplicitAnnotatedFieldSchema, type UnannotateImplicitFieldSchema } from "./simple-tree/index.js";
/**
 * Not intended for use outside of this package.
 *
 * @privateRemarks
 * This namespace is a collection of internal system types relate to {@link TableSchema}.
 * This namespace should be strictly type-exported by the package.
 * All members should be tagged with `@system`.
 *
 * @system @internal
 */
export declare namespace System_TableSchema {
    /**
     * A base interface for factory input options which include an schema factory.
     * @remarks This interface should not be referenced directly.
     * @privateRemarks This interface primarily exists to provide a single home for property documentation.
     * @system @internal
     */
    interface OptionsWithSchemaFactory<TScope extends string | undefined = string | undefined> {
        /**
         * Schema factory with which the Column schema will be associated.
         * @remarks Can be used to associate the resulting schema with an existing {@link SchemaFactory.scope|scope}.
         */
        readonly schemaFactory: SchemaFactoryAlpha<TScope>;
    }
    /**
     * A base interface for factory input options which include the table cell schema.
     * @remarks This interface should not be referenced directly.
     * @privateRemarks This interface primarily exists to provide a single home for property documentation.
     * @system @internal
     */
    interface OptionsWithCellSchema<TCellSchema extends ImplicitAllowedTypes = ImplicitAllowedTypes> {
        /**
         * Schema for the table's cells.
         */
        readonly cell: TCellSchema;
    }
    /**
     * Base options for creating table cow schema.
     * @remarks Includes parameters common to all column factory overloads.
     * @system @internal
     */
    type CreateColumnOptionsBase<TInputScope extends string | undefined = string | undefined> = OptionsWithSchemaFactory<TInputScope>;
    /**
     * Factory for creating new table column schema.
     * @system @internal
     */
    function createColumnInternal<const TInputScope extends string | undefined, const TPropsSchema extends ImplicitAnnotatedFieldSchema>(inputSchemaFactory: SchemaFactoryAlpha<TInputScope>, propsSchema: TPropsSchema): TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Column">, NodeKind.Object, TreeNode & TableSchema.IColumn<TPropsSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Column">, NodeKind, unknown>, object & {
        readonly id?: string | undefined;
    } & (FieldHasDefault<UnannotateImplicitFieldSchema<TPropsSchema>> extends true ? {
        /**
         * The column's properties.
         * @remarks This is a user-defined schema that can be used to store additional information about the column.
         */
        props?: InsertableTreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TPropsSchema>> | undefined;
    } : {
        /**
         * The column's properties.
         * @remarks This is a user-defined schema that can be used to store additional information about the column.
         */
        props: InsertableTreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TPropsSchema>>;
    }), true, {
        readonly props: TPropsSchema;
        readonly id: FieldSchema<FieldKind.Identifier, import("./simple-tree/leafNodeSchema.js").LeafSchema<"string", string>, unknown>;
    }>;
    /**
     * Base column schema type.
     * @sealed @system @internal
     */
    type ColumnSchemaBase<TScope extends string | undefined = string | undefined, TPropsSchema extends ImplicitFieldSchema = ImplicitFieldSchema> = ReturnType<typeof TableSchema.createColumn<TScope, TPropsSchema>>;
    /**
     * Base options for creating table row schema.
     * @remarks Includes parameters common to all row factory overloads.
     * @system @internal
     */
    type CreateRowOptionsBase<TScope extends string | undefined = string | undefined, TCell extends ImplicitAllowedTypes = ImplicitAllowedTypes> = OptionsWithSchemaFactory<TScope> & OptionsWithCellSchema<TCell>;
    /**
     * Factory for creating new table row schema.
     *
     * @sealed @internal
     */
    function createRowInternal<const TInputScope extends string | undefined, const TCellSchema extends ImplicitAllowedTypes, const TPropsSchema extends ImplicitFieldSchema>(inputSchemaFactory: SchemaFactoryAlpha<TInputScope>, cellSchema: TCellSchema, propsSchema: TPropsSchema): TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row">, NodeKind.Object, TreeNode & TableSchema.IRow<TCellSchema, TPropsSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row">, NodeKind, unknown>, object & {
        readonly id?: string | undefined;
        readonly cells: (import("./simple-tree/schemaTypes.js").InsertableTypedNode<TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, import("./simple-tree/mapNode.js").TreeMapNode<TCellSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, unknown>, import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema>, true, TCellSchema, undefined>, import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, true, TCellSchema, import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema>, unknown> & (new (data?: InternalTreeNode | import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema> | undefined) => import("./simple-tree/mapNode.js").TreeMapNode<TCellSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, unknown>)> | undefined) & import("./simple-tree/schemaTypes.js").InsertableTypedNode<TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, import("./simple-tree/mapNode.js").TreeMapNode<TCellSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, unknown>, import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema>, true, TCellSchema, undefined>, import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, true, TCellSchema, import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema>, unknown> & (new (data?: InternalTreeNode | import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema> | undefined) => import("./simple-tree/mapNode.js").TreeMapNode<TCellSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, unknown>)>;
    } & (FieldHasDefault<TPropsSchema> extends true ? {
        /**
         * The row's properties.
         * @remarks This is a user-defined schema that can be used to store additional information
         * about the row.
         */
        props?: InsertableTreeFieldFromImplicitField<TPropsSchema> | undefined;
    } : {
        /**
         * The row's properties.
         * @remarks This is a user-defined schema that can be used to store additional information
         * about the row.
         */
        props: InsertableTreeFieldFromImplicitField<TPropsSchema>;
    }), true, {
        readonly props: TPropsSchema;
        readonly id: FieldSchema<FieldKind.Identifier, import("./simple-tree/leafNodeSchema.js").LeafSchema<"string", string>, unknown>;
        readonly cells: FieldSchema<FieldKind.Required, TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, import("./simple-tree/mapNode.js").TreeMapNode<TCellSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Row.cells">, NodeKind.Map, unknown>, import("./simple-tree/mapNode.js").MapNodeInsertableData<TCellSchema>, true, TCellSchema, undefined>, unknown>;
    }>;
    /**
     * Base row schema type.
     * @sealed @system @internal
     */
    type RowSchemaBase<TScope extends string | undefined = string | undefined, TCellSchema extends ImplicitAllowedTypes = ImplicitAllowedTypes, TPropsSchema extends ImplicitFieldSchema = ImplicitFieldSchema> = ReturnType<typeof TableSchema.createRow<TScope, TCellSchema, TPropsSchema>>;
    /**
     * Base options for creating table schema.
     * @remarks Includes parameters common to all table factory overloads.
     * @system @internal
     */
    type TableFactoryOptionsBase<TScope extends string | undefined = string | undefined, TCell extends ImplicitAllowedTypes = ImplicitAllowedTypes> = OptionsWithSchemaFactory<TScope> & OptionsWithCellSchema<TCell>;
    /**
     * Factory for creating new table schema.
     * @system @internal
     */
    function createTableInternal<const TInputScope extends string | undefined, const TCellSchema extends ImplicitAllowedTypes, const TColumnSchema extends ColumnSchemaBase<TInputScope> = ColumnSchemaBase<TInputScope>, const TRowSchema extends RowSchemaBase<TInputScope, TCellSchema> = RowSchemaBase<TInputScope, TCellSchema>>(inputSchemaFactory: SchemaFactoryAlpha<TInputScope>, _cellSchema: TCellSchema, columnSchema: TColumnSchema, rowSchema: TRowSchema): TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table">, NodeKind.Object, TreeNode & TableSchema.ITable<TCellSchema, TColumnSchema, TRowSchema> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table">, NodeKind, unknown>, object & {
        readonly rows: (import("./simple-tree/schemaTypes.js").InsertableTypedNode<import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, true, TRowSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>> | undefined) => TreeArrayNode<TRowSchema, [TRowSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TRowSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TRowSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, unknown>), import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, true, TRowSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>> | undefined) => TreeArrayNode<TRowSchema, [TRowSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TRowSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TRowSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, unknown>)> | undefined) & import("./simple-tree/schemaTypes.js").InsertableTypedNode<import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, true, TRowSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>> | undefined) => TreeArrayNode<TRowSchema, [TRowSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TRowSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TRowSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, unknown>), import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, true, TRowSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>> | undefined) => TreeArrayNode<TRowSchema, [TRowSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TRowSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TRowSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, unknown>)>;
        readonly columns: (import("./simple-tree/schemaTypes.js").InsertableTypedNode<import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, true, TColumnSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>> | undefined) => TreeArrayNode<TColumnSchema, [TColumnSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TColumnSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TColumnSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, unknown>), import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, true, TColumnSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>> | undefined) => TreeArrayNode<TColumnSchema, [TColumnSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TColumnSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TColumnSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, unknown>)> | undefined) & import("./simple-tree/schemaTypes.js").InsertableTypedNode<import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, true, TColumnSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>> | undefined) => TreeArrayNode<TColumnSchema, [TColumnSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TColumnSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TColumnSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, unknown>), import("./simple-tree/index.js").TreeNodeSchemaCore<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, true, TColumnSchema, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>>, unknown> & (new (data?: InternalTreeNode | Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>> | undefined) => TreeArrayNode<TColumnSchema, [TColumnSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TColumnSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TColumnSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, unknown>)>;
    }, true, {
        readonly rows: TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, TreeArrayNode<TRowSchema, [TRowSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TRowSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TRowSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.rows">, NodeKind.Array, unknown>, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TRowSchema>>, true, TRowSchema, undefined>;
        readonly columns: TreeNodeSchemaClass<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, TreeArrayNode<TColumnSchema, [TColumnSchema] extends [ImplicitAllowedTypes] ? TreeNodeFromImplicitAllowedTypes<TColumnSchema> : TreeNodeFromImplicitAllowedTypes<ImplicitAllowedTypes>, [TColumnSchema] extends [ImplicitAllowedTypes] ? InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema> : never, import("./simple-tree/arrayNode.js").ReadonlyArrayNode<TreeNode | import("./simple-tree/schemaTypes.js").TreeLeafValue>> & WithType<ScopedSchemaName<ScopedSchemaName<TInputScope, "table">, "Table.columns">, NodeKind.Array, unknown>, Iterable<InsertableTreeNodeFromImplicitAllowedTypes<TColumnSchema>>, true, TColumnSchema, undefined>;
    }>;
    /**
     * Base row schema type.
     * @sealed @system @internal
     */
    type TableSchemaBase<TScope extends string | undefined, TCell extends ImplicitAllowedTypes, TColumn extends ColumnSchemaBase<TScope> = ColumnSchemaBase<TScope>, TRow extends RowSchemaBase<TScope, TCell, ImplicitAllowedTypes> = RowSchemaBase<TScope, TCell, ImplicitAllowedTypes>> = ReturnType<typeof TableSchema.createTable<TScope, TCell, TColumn, TRow>>;
}
/**
 * Contains types and factories for creating schema to represent dynamic tabular data.
 * @privateRemarks TODO: document in more detail and add `@example`s.
 * @internal
 */
export declare namespace TableSchema {
    /**
     * A column in a table.
     * @remarks Implemented by the schema class returned from {@link TableSchema.(createColumn:2)}.
     * @sealed @internal
     */
    interface IColumn<TProps extends ImplicitAnnotatedFieldSchema = ImplicitAnnotatedFieldSchema> {
        /**
         * The unique identifier of the column.
         * @remarks Uniquely identifies the node within the entire tree, not just the table.
         */
        readonly id: string;
        /**
         * The column's properties.
         * @remarks This is a user-defined schema that can be used to store additional information about the column.
         * @privateRemarks
         * Note: these docs are duplicated on the inline type definitions in {@link createColumn}.
         * If you update the docs here, please also update the inline type definitions.
         */
        get props(): TreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TProps>> | undefined;
        set props(value: InsertableTreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TProps>>);
    }
    /**
     * Factory for creating new table column schema.
     * @internal
     */
    function createColumn<const TScope extends string | undefined>({ schemaFactory, }: System_TableSchema.CreateColumnOptionsBase<TScope>): ReturnType<typeof System_TableSchema.createColumnInternal<TScope, FieldSchema<FieldKind.Optional, typeof SchemaFactoryAlpha.null>>>;
    /**
     * Factory for creating new table column schema.
     * @internal
     */
    function createColumn<const TScope extends string | undefined, const TProps extends ImplicitFieldSchema>({ schemaFactory, props, }: System_TableSchema.CreateColumnOptionsBase<TScope> & {
        /**
         * Optional column properties.
         */
        readonly props: TProps;
    }): ReturnType<typeof System_TableSchema.createColumnInternal<TScope, TProps>>;
    /**
     * A row in a table.
     * @remarks Implemented by the schema class returned from {@link TableSchema.(createRow:2)}.
     * @sealed @internal
     */
    interface IRow<TCell extends ImplicitAllowedTypes, TProps extends ImplicitAnnotatedFieldSchema = ImplicitAnnotatedFieldSchema> {
        /**
         * The unique identifier of the row.
         * @remarks Uniquely identifies the node within the entire tree, not just the table.
         */
        readonly id: string;
        /**
         * Gets the cell in the specified column.
         * @returns The cell if it exists, otherwise undefined.
         */
        getCell(column: IColumn): TreeNodeFromImplicitAllowedTypes<TCell> | undefined;
        /**
         * Gets the cell in the specified column, denoted by column ID.
         * @returns The cell if it exists, otherwise undefined.
         */
        getCell(columnId: string): TreeNodeFromImplicitAllowedTypes<TCell> | undefined;
        /**
         * Sets the cell in the specified column.
         * @remarks To remove a cell, call {@link TableSchema.IRow.(removeCell:1)} instead.
         */
        setCell(column: IColumn, value: InsertableTreeNodeFromImplicitAllowedTypes<TCell>): void;
        /**
         * Sets the cell in the specified column, denoted by column ID.
         * @remarks To remove a cell, call {@link TableSchema.IRow.(removeCell:2)} instead.
         */
        setCell(columnId: string, value: InsertableTreeNodeFromImplicitAllowedTypes<TCell>): void;
        /**
         * Removes the cell in the specified column.
         * @privateRemarks TODO: return removed cell
         */
        removeCell(column: IColumn): void;
        /**
         * Removes the cell in the specified column, denoted by column ID.
         * @privateRemarks TODO: return removed cell
         */
        removeCell(columnId: string): void;
        /**
         * The row's properties.
         * @remarks This is a user-defined schema that can be used to store additional information about the row.
         * @privateRemarks
         * Note: these docs are duplicated on the inline type definitions in {@link createColumn}.
         * If you update the docs here, please also update the inline type definitions.
         */
        get props(): TreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TProps>>;
        set props(value: InsertableTreeFieldFromImplicitField<UnannotateImplicitFieldSchema<TProps>>);
    }
    /**
     * Factory for creating new table column schema.
     * @internal
     */
    function createRow<const TScope extends string | undefined, const TCell extends ImplicitAllowedTypes>({ schemaFactory, cell, }: System_TableSchema.CreateRowOptionsBase<TScope, TCell>): ReturnType<typeof System_TableSchema.createRowInternal<TScope, TCell, FieldSchema<FieldKind.Optional, typeof SchemaFactoryAlpha.null>>>;
    /**
     * Factory for creating new table column schema.
     * @internal
     */
    function createRow<const TScope extends string | undefined, const TCell extends ImplicitAllowedTypes, const TProps extends ImplicitFieldSchema>({ schemaFactory, cell, props, }: System_TableSchema.CreateRowOptionsBase<TScope, TCell> & {
        /**
         * Optional row properties.
         */
        readonly props: TProps;
    }): ReturnType<typeof System_TableSchema.createRowInternal<TScope, TCell, TProps>>;
    /**
     * A key to uniquely identify a cell in a table.
     * @internal
     */
    interface CellKey {
        /**
         * {@link TableSchema.IColumn.id} of the containing {@link TableSchema.IColumn}.
         */
        readonly columnId: string;
        /**
         * {@link TableSchema.IRow.id} of the containing {@link TableSchema.IRow}.
         */
        readonly rowId: string;
    }
    /**
     * {@link TableSchema.ITable.insertColumn} parameters.
     * @internal
     */
    interface InsertColumnParameters<TInsertableColumn> {
        /**
         * The index at which to insert the new column.
         * @remarks If not provided, the column will be appended to the end of the table.
         */
        readonly index?: number | undefined;
        /**
         * The column to insert.
         */
        readonly column: TInsertableColumn;
    }
    /**
     * {@link TableSchema.ITable.insertRows} parameters.
     * @internal
     */
    interface InsertRowsParameters<TInsertableRow> {
        /**
         * The index at which to insert the new rows.
         * @remarks If not provided, the rows will be appended to the end of the table.
         */
        readonly index?: number | undefined;
        /**
         * The rows to insert.
         */
        readonly rows: TInsertableRow[];
    }
    /**
     * {@link TableSchema.ITable.setCell} parameters.
     * @internal
     */
    interface SetCellParameters<TInsertableCell> {
        /**
         * The key to uniquely identify a cell in a table.
         */
        readonly key: CellKey;
        /**
         * The cell to set.
         */
        readonly cell: TInsertableCell;
    }
    /**
     * A table.
     * @sealed @internal
     */
    interface ITable<TCell extends ImplicitAllowedTypes, TColumn extends ImplicitAllowedTypes, TRow extends ImplicitAllowedTypes> {
        /**
         * The table's columns.
         */
        readonly columns: TreeArrayNode<TColumn>;
        /**
         * The table's rows.
         */
        readonly rows: TreeArrayNode<TRow>;
        /**
         * Gets a table column by its {@link TableSchema.IRow.id}.
         */
        getColumn(id: string): TreeNodeFromImplicitAllowedTypes<TColumn> | undefined;
        /**
         * Gets a table row by its {@link TableSchema.IRow.id}.
         */
        getRow(id: string): TreeNodeFromImplicitAllowedTypes<TRow> | undefined;
        /**
         * Gets a cell in the table by column and row IDs.
         * @param key - A key that uniquely distinguishes a cell in the table, represented as a combination of the column ID and row ID.
         * @privateRemarks TODO: add overload that takes row and column nodes.
         */
        getCell(key: CellKey): TreeNodeFromImplicitAllowedTypes<TCell> | undefined;
        /**
         * Inserts a column into the table.
         * @throws Throws an error if the column is already in the tree, or if the specified index is out of range.
         */
        insertColumn(params: InsertColumnParameters<InsertableTreeNodeFromImplicitAllowedTypes<TColumn>>): TreeNodeFromImplicitAllowedTypes<TColumn>;
        /**
         * Inserts 0 or more rows into the table.
         * @throws Throws an error if any of the rows are already in the tree, or if the specified index is out of range.
         */
        insertRows(params: InsertRowsParameters<InsertableTreeNodeFromImplicitAllowedTypes<TRow>>): TreeNodeFromImplicitAllowedTypes<TRow>[];
        /**
         * Sets the cell at the specified location in the table.
         * @remarks To remove a cell, call {@link TableSchema.ITable.removeCell} instead.
         * @privateRemarks TODO: add overload that takes column/row nodes?
         */
        setCell(params: SetCellParameters<InsertableTreeNodeFromImplicitAllowedTypes<TCell>>): void;
        /**
         * Removes the specified column from the table.
         * @remarks Note: this does not remove any cells from the table's rows.
         * @privateRemarks
         * TODO:
         * - Policy for when the column is not in the table.
         * - Actually remove corresponding cells from table rows.
         */
        removeColumn: (column: TreeNodeFromImplicitAllowedTypes<TColumn>) => void;
        /**
         * Removes 0 or more rows from the table.
         * @privateRemarks TODO: policy for when 1 or more rows are not in the table.
         */
        removeRows: (rows: readonly TreeNodeFromImplicitAllowedTypes<TRow>[]) => void;
        /**
         * Removes all rows from the table.
         */
        removeAllRows: () => void;
        /**
         * Removes the cell at the specified location in the table.
         * @privateRemarks TODO: add overload that takes column/row nodes?
         */
        removeCell: (key: CellKey) => void;
    }
    /**
     * Factory for creating new table schema without specifying row or column schema.
     * @internal
     */
    function createTable<const TScope extends string | undefined, const TCell extends ImplicitAllowedTypes>({ schemaFactory, cell, }: System_TableSchema.TableFactoryOptionsBase<TScope, TCell>): ReturnType<typeof System_TableSchema.createTableInternal<TScope, TCell>>;
    /**
     * Factory for creating new table schema without specifying row schema.
     * @internal
     */
    function createTable<const TScope extends string | undefined, const TCell extends ImplicitAllowedTypes, const TColumn extends System_TableSchema.ColumnSchemaBase<TScope>>({ schemaFactory, cell, column, }: System_TableSchema.TableFactoryOptionsBase<TScope, TCell> & {
        readonly column: TColumn;
    }): ReturnType<typeof System_TableSchema.createTableInternal<TScope, TCell, TColumn>>;
    /**
     * Factory for creating new table schema.
     * @internal
     */
    function createTable<const TScope extends string | undefined, const TCell extends ImplicitAllowedTypes, const TColumn extends System_TableSchema.ColumnSchemaBase<TScope>, const TRow extends System_TableSchema.RowSchemaBase<TScope, TCell>>({ schemaFactory, cell, column, row, }: System_TableSchema.TableFactoryOptionsBase<TScope, TCell> & {
        readonly column: TColumn;
        readonly row: TRow;
    }): ReturnType<typeof System_TableSchema.createTableInternal<TScope, TCell, TColumn, TRow>>;
}
//# sourceMappingURL=tableSchema.d.ts.map